#!/usr/bin/python
# -*- coding: utf-8 -*-
# JSON Grid dumper
# (C) 2018 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si:

from __future__ import unicode_literals

from .datatypes import Quantity, Coordinate, Ref, Bin, Uri, \
        MARKER
from .zoneinfo import timezone_name
from .jsonparser import MARKER_STR
from .version import LATEST_VER, VER_3_0

import datetime
import functools
import json
import six

def dump_grid(grid):
    return json.dumps({
        'meta': dump_meta(grid.metadata, version=grid.version, grid=True),
        'cols': dump_columns(grid.column, version=grid.version),
        'rows': dump_rows(grid),
    })

def dump_meta(meta, version=LATEST_VER, grid=False):
    _dump = functools.partial(dump_meta_item, version=version)
    _meta = dict(map(_dump, list(meta.items())))
    if grid:
        _meta['ver'] = str(version)
    return _meta

def dump_meta_item(item, version=LATEST_VER):
    (item_id, item_value) = item
    return (dump_id(item_id, version=version), \
            dump_scalar(item_value, version=version))

def dump_columns(cols, version=LATEST_VER):
    _dump = functools.partial(dump_column, version=version)
    _cols = list(zip(*list(cols.items())))
    return list(map(_dump, *_cols))

def dump_column(col, col_meta, version=LATEST_VER):
    if bool(col_meta):
        _meta = dump_meta(col_meta, version=version)
    else:
        _meta = {}
    _meta['name'] = col
    return _meta

def dump_rows(grid):
    return list(map(functools.partial(dump_row, grid), grid))

def dump_row(grid, row):
    return dict([
        (c, dump_scalar(row.get(c), version=grid.version))
        for c in list(grid.column.keys())])

def dump_scalar(scalar, version=LATEST_VER):
    if scalar is None:
        return None
    elif scalar is MARKER:
        return MARKER_STR
    elif isinstance(scalar, list):
        return dump_list(scalar, version=version)
    elif isinstance(scalar, bool):
        return dump_bool(scalar, version=version)
    elif isinstance(scalar, Ref):
        return dump_ref(scalar, version=version)
    elif isinstance(scalar, Bin):
        return dump_bin(scalar, version=version)
    elif isinstance(scalar, Uri):
        return dump_uri(scalar, version=version)
    elif isinstance(scalar, six.string_types):
        return dump_str(scalar, version=version)
    elif isinstance(scalar, datetime.datetime):
        return dump_date_time(scalar, version=version)
    elif isinstance(scalar, datetime.time):
        return dump_time(scalar, version=version)
    elif isinstance(scalar, datetime.date):
        return dump_date(scalar, version=version)
    elif isinstance(scalar, Coordinate):
        return dump_coord(scalar, version=version)
    elif isinstance(scalar, Quantity):
        return dump_quantity(scalar, version=version)
    elif isinstance(scalar, float) or \
            isinstance(scalar, int) or \
            isinstance(scalar, int):
        return dump_decimal(scalar, version=version)
    else: # pragma: no cover
        raise NotImplementedError('Unhandled case: %r' % scalar)

def dump_id(id_str, version=LATEST_VER):
    return id_str

def dump_str(str_value, version=LATEST_VER):
    return u's:%s' % str_value

def dump_uri(uri_value, version=LATEST_VER):
    return u'u:%s' % uri_value

def dump_bin(bin_value, version=LATEST_VER):
    return u'b:%s' % bin_value

def dump_quantity(quantity, version=LATEST_VER):
    if (quantity.unit is None) or (quantity.unit == ''):
        return dump_decimal(quantity.value, version=version)
    else:
        return 'n:%f %s' % (quantity.value, quantity.unit)

def dump_decimal(decimal, version=LATEST_VER):
    return 'n:%f' % decimal

def dump_bool(bool_value, version=LATEST_VER):
    return bool_value

def dump_coord(coordinate, version=LATEST_VER):
    return 'c:%f,%f' % (coordinate.latitude, coordinate.longitude)

def dump_ref(ref, version=LATEST_VER):
    if ref.has_value:
        return u'r:%s %s' % (ref.name, ref.value)
    else:
        return u'r:%s' % ref.name

def dump_date(date, version=LATEST_VER):
    return 'd:%s' % date.isoformat()

def dump_time(time, version=LATEST_VER):
    return 'h:%s' % time.isoformat()

def dump_date_time(date_time, version=LATEST_VER):
    tz_name = timezone_name(date_time, version=version)
    return 't:%s %s' % (date_time.isoformat(), tz_name)

def dump_list(lst, version=LATEST_VER):
    if version < VER_3_0:
        raise ValueError('Project Haystack %s '\
                'does not support lists' % version)
    return list(map(functools.partial(dump_scalar, version=version), lst))



import random
from io import open

_RAW_UNIT_DATA = None

def get_units():
    global _RAW_UNIT_DATA
    if _RAW_UNIT_DATA is None:
        unit_data = {}

        for raw_row in open(
                'tests/project_haystack_units.txt',
                mode='r', encoding='utf-8'):
            row = raw_row.split(u',')
            if not bool(row):
                continue
            if row[0].startswith('--'):
                continue

            if len(row) == 1:
                unit_data[row[0]] = row[0]
            else:
                unit_data[row[0]] = row[1]
        _RAW_UNIT_DATA = unit_data
    return _RAW_UNIT_DATA.copy()

def get_random_unit():
    return random.choice(list(get_units().values()))



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Sortable dict helper class
# (C) 2016 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si:

from collections import MutableMapping

class SortableDict(MutableMapping):
    """
    A dict-like object that permits value ordering/re-ordering.
    """

    def __init__(self, initial=None, validate_fn=None):
        self._values = {}
        self._order = []
        self._validate_fn = validate_fn
        super(SortableDict, self).__init__()

        # Copy initial values into dict.
        if initial is not None:
            # If we're given a dict; make it a list of items
            if isinstance(initial, dict):
                initial = list(initial.items())

            for (key, val) in initial:
                self[key] = val

    def __repr__(self):
        return '%s{%s}' % (self.__class__.__name__,
                ', '.join([
                    '%r=%r' % (k,v) for k,v in list(self.items())
                ]))

    def __getitem__(self, key):
        return self._values[key]

    def __setitem__(self, key, value):
        return self.add_item(key, value)

    def __delitem__(self, key):
        del self._values[key]
        self._order.remove(key)

    def __iter__(self):
        return iter(self._order)

    def __len__(self):
        return len(self._order)

    def add_item(self, key, value, after=False, index=None, pos_key=None,
            replace=True):
        """
        Add an item at a specific location, possibly replacing the
        existing item.

        If after is True, we insert *after* the given index, otherwise we
        insert before.

        The position is specified using either index or pos_key, the former
        specifies the position from the start of the array (base 0).  pos_key
        specifies the name of another key, and positions the new key relative
        to that key.

        When replacing, the position will be left un-changed unless a location
        is specified explicitly.
        """
        if self._validate_fn:
            self._validate_fn(value)

        if (index is not None) and (pos_key is not None):
            raise ValueError('Either specify index or pos_key, not both.')
        elif pos_key is not None:
            try:
                index = self.index(pos_key)
            except ValueError:
                raise KeyError('%r not found' % pos_key)

        if after and (index is not None):
            # insert inserts *before* index, so increment by one.
            index += 1

        if key in self._values:
            if not replace:
                raise KeyError('%r is duplicate' % key)

            if index is not None:
                # We are re-locating.
                del self[key]
            else:
                # We are updating
                self._values[key] = value
                return

        if index is not None:
            # Place at given position
            self._order.insert(index, key)
        else:
            # Place at end
            self._order.append(key)
        self._values[key] = value

    def at(self, index):
        """
        Return the key at the given index.
        """
        return self._order[index]

    def value_at(self, index):
        """
        Return the value at the given index.
        """
        return self[self.at(index)]

    def index(self, *args, **kwargs):
        return self._order.index(*args, **kwargs)

    def reverse(self, *args, **kwargs):
        return self._order.reverse(*args, **kwargs)

    def sort(self, *args, **kwargs):
        return self._order.sort(*args, **kwargs)

    def pop_at(self, index):
        """
        Remove the key at the given index and return its value.
        """
        return self.pop(self.at(index))



#!/usr/bin/python
from setuptools import setup
import sys
from hszinc import __version__

requirements = [
            'pyparsing',
            'pytz',
            'iso8601',
            'six',
            'pint',
]

setup(name='hszinc',
        url='https://github.com/vrtsystems/hszinc',
        description='Parser and dumper for Project '\
                    'Haystack ZINC (Zinc is not CSV) file format',
        version=__version__,
        author='VRT Systems',
        author_email='support@vrt.com.au',
        license='BSD',
        packages=[
            'hszinc',
        ],
        classifiers=[
            'Development Status :: 4 - Beta',
            'Environment :: No Input/Output (Daemon)',
            'Environment :: Other Environment',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: BSD License',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Topic :: Scientific/Engineering',
            'Topic :: Scientific/Engineering :: Information Analysis',
            'Topic :: Software Development :: Libraries :: Python Modules',
        ],
        requires=requirements,
        install_requires=requirements
)



# -*- coding: utf-8 -*-
# Zinc dumping and parsing module
# (C) 2016 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si:

import warnings

# First verify if pint is available
PINT_AVAILABLE = False
try:
    from pint import UnitRegistry
    PINT_AVAILABLE = True
    from .pintutil import define_haystack_units
    ureg = define_haystack_units()
except ImportError: # pragma: no cover
    # For setup.py to interrogate the version information.  This should *NOT*
    # get executed in production, and if it did, things wouldn't work anyway.
    ureg = {'Quantity': None}

try:
    from .grid import Grid
    from .dumper import dump, dump_scalar
    from .parser import parse, parse_scalar, MODE_JSON, MODE_ZINC
    from .metadata import MetadataObject
    from .datatypes import Quantity, Coordinate, Uri, Bin, MARKER, REMOVE, Ref, use_pint
    from .version import Version, VER_2_0, VER_3_0, LATEST_VER
    Q_ = Quantity
    __all__ = ['Grid', 'dump', 'parse', 'dump_scalar', 'parse_scalar',
            'MetadataObject', 'ureg',
            'Coordinate', 'Uri', 'Bin', 'MARKER', 'REMOVE', 'Ref',
            'MODE_JSON', 'MODE_ZINC',
            'VER_2_0', 'VER_3_0', 'LATEST_VER', 'Version']
except ImportError as e: # pragma: no cover
    # For setup.py to interrogate the version information.  This should *NOT*
    # get executed in production, and if it did, things wouldn't work anyway.
    warnings.warn(
        'Failed to import libraries: %s, dependencies may be missing' \
                % e)

__author__ = 'VRT Systems'
__copyright__ = 'Copyright 2016, VRT Systems'
__credits__ = ['VRT Systems']
__license__ = 'BSD'
__version__ = '1.2.2'
__maintainer__ = 'VRT Systems'
__email__ = 'support@vrt.com.au'



# -*- coding: utf-8 -*-
# Version handling/parsing tests
# (C) 2018 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si:

# Assume unicode literals as per Python 3
from __future__ import unicode_literals

from hszinc import Version
import warnings
import re

def test_malformed():
    try:
        Version('this is not a valid string')
        assert False, 'Accepted a bogus version string'
    except ValueError:
        pass

def test_str():
    assert str(Version('1.0')) == '1.0'
    assert str(Version('1.0a')) == '1.0a'

def test_str_cmp():
    assert Version('1.0') == '1.0'

def test_version_eq():
    assert Version('1') == Version('1.0')
    assert Version('1.0') == Version('1.0.0')
    assert Version('1.0a') == Version('1.0.0a')

def test_version_ne():
    assert Version('1.1') != Version('2.2')
    assert Version('1.0a') != Version('1.0')

def test_version_lt():
    assert Version('1.0') < Version('1.1')
    assert Version('1.9') < Version('1.10')
    assert Version('1.0') < Version('1.0a')
    assert Version('1.0a') < Version('1.0b')

def test_version_le():
    assert Version('1.0') <= Version('1.0')
    assert Version('1.0') <= Version('1.5')

def test_version_ge():
    assert Version('1.0') >= Version('1.0')
    assert Version('1.0') >= Version('0.9')

def test_version_gt():
    assert Version('1.0') > Version('0.9')
    assert Version('1.0a') > Version('1.0')
    assert Version('1.0b') > Version('1.0a')

_WARNING_RE = re.compile(
            r'This version of hszinc does not yet support version ([\d\.]+), '\
            r'please seek a newer version or file a bug.  Closest '\
            r'\((older|newer)\) version supported is ([\d\.]+).')

def _check_warning(w):
    assert issubclass(w.category, UserWarning)

    warning_match = _WARNING_RE.match(str(w.message))
    assert warning_match is not None
    (detect_ver_str, older_newer, used_ver_str) = warning_match.groups()

    return (older_newer, Version(detect_ver_str),
            Version(used_ver_str))

def test_unsupported_old():
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        assert Version.nearest("1.0") == Version("2.0")

        # Check we got a warning for that old crusty version.
        assert len(w) == 1
        (older_newer, detect_ver, used_ver) = _check_warning(w[-1])
        assert older_newer == 'newer'
        assert used_ver == Version('2.0')

def test_unsupported_newer():
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        assert Version.nearest("2.5") == Version("3.0")

        # Check we got a warning for that oddball newer version.
        assert len(w) == 1
        (older_newer, detect_ver, used_ver) = _check_warning(w[-1])
        assert older_newer == 'newer'
        assert used_ver == Version('3.0')

def test_unsupported_bleedingedge():
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        assert Version.nearest("9999.9999") == Version("3.0")

        # Check we got a warning for that bleeding edge version.
        assert len(w) == 1
        (older_newer, detect_ver, used_ver) = _check_warning(w[-1])
        assert older_newer == 'older'
        assert used_ver == Version('3.0')



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Zinc Grid Metadata
# (C) 2016 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si:

from .zincparser import parse_grid as parse_zinc_grid, \
                        parse_scalar as parse_zinc_scalar
from .jsonparser import parse_grid as parse_json_grid, \
                        parse_scalar as parse_json_scalar
import re
import six
import functools
import json

# Bring in version handling
from .version import Version, LATEST_VER

GRID_SEP = re.compile(r'\n\n+')

MODE_ZINC = 'zinc'
MODE_JSON = 'json'

def parse(grid_str, mode=MODE_ZINC, charset='utf-8'):
    '''
    Parse the given Zinc text and return the equivalent data.
    '''
    # Decode incoming text (or python3 will whine!)
    if isinstance(grid_str, six.binary_type):
        grid_str = grid_str.decode(encoding=charset)

    # Split the separate grids up, the grammar definition has trouble splitting
    # them up normally.  This will truncate the newline off the end of the last
    # row.
    _parse = functools.partial(parse_grid, mode=mode,
            charset=charset)
    if mode == MODE_JSON:
        if isinstance(grid_str, six.string_types):
            grid_data = json.loads(grid_str)
        else:
            grid_data = grid_str
        if isinstance(grid_data, dict):
            return _parse(grid_data)
        else:
            return list(map(_parse, grid_data))
    else:
        return list(map(_parse, GRID_SEP.split(grid_str.rstrip())))

def parse_grid(grid_str, mode=MODE_ZINC, charset='utf-8'):
    # Decode incoming text
    if isinstance(grid_str, six.binary_type): # pragma: no cover
        # No coverage here, because it *should* be handled above unless the user
        # is pre-empting us by calling `parse_grid` directly.
        grid_str = grid_str.decode(encoding=charset)

    if mode == MODE_ZINC:
        return parse_zinc_grid(grid_str)
    elif mode == MODE_JSON:
        return parse_json_grid(grid_str)
    else: # pragma: no cover
        raise NotImplementedError('Format not implemented: %s' % mode)

def parse_scalar(scalar, mode=MODE_ZINC, version=LATEST_VER, charset='utf-8'):
    # Decode version string
    if not isinstance(version, Version):
        version = Version(version)

    # Decode incoming text
    if isinstance(scalar, six.binary_type):
        scalar = scalar.decode(encoding=charset)

    if mode == MODE_ZINC:
        return parse_zinc_scalar(scalar, version=version)
    elif mode == MODE_JSON:
        return parse_json_scalar(scalar, version=version)
    else: # pragma: no cover
        raise NotImplementedError('Format not implemented: %s' % mode)
